/* The serialized file format is pretty generic... each line (using any line
 * separator, so we don't mind being moved between platforms) consists of
 * a command name, and some parameters (optionally). The commands 'start'
 * and 'end' mark the chunks of properties for each object - in this case
 * motifs. Every command inside a start/end block is considered a property
 * for the object. There are some rules, but we are generally pretty flexible.
 *
 * Example file:
 *   START
 *     TITLE Default
 *     KEY default
 *     URL chrome://chatzilla/skin/output-default.css
 *   END
 *   START
 *     TITLE My own motif
 *     KEY mine
 *     URL file:///C:/Documents%20and%20Settings/User/My%20Documents/ChatZilla-motif.css
 *   END
 *
 * The whitespace at the start of the inner lines is generated by the
 * serialisation process, but is ignored when parsing - it is only to make
 * the file more readable.
 */

function TextSerializer(file)
{
    this._initialized = false;
    if (typeof file == "string")
        this._file = new nsLocalFile(file);
    else
        this._file = file;
    this._open = false;
    this._buffer = "";
    this._lines = [];
    this.lineEnd = "\n";
    this._initialized = true;
}

/* open(direction)
 *
 * Opens the serializer on the file specified when created, in either the read
 * ("<") or write (">") directions. When the file is open, only the appropriate
 * direction of serialization/deserialization may be performed.
 *
 * Note: serialize and deserialize automatically open the file is it is not
 *       open.
 */
TextSerializer.prototype.open =
function ts_open(dir)
{
    if (!ASSERT((dir == ">") || (dir == "<"), "Bad serilization direction!"))
        return false;
    if (this._open)
        return false;

    this._fileStream = new LocalFile(this._file, dir);
    if ((typeof this._fileStream == "object") && this._fileStream)
        this._open = true;

    return this._open;
}

/* close()
 *
 * Closes the file steam and ends reading or writing.
 */
TextSerializer.prototype.close =
function ts_close()
{
    if (this._open)
    {
        this._fileStream.close();
        delete this._fileStream;
        this._open = false;
    }
    return !this._open;
}

/* serialize(object)
 *
 * Serializes a single object into the file stream. All properties of the object
 * are stored in the stream, include properties that contain other objects.
 */
TextSerializer.prototype.serialize =
function ts_serialize(obj)
{
    if (!this._open)
        this.open(">");
    if (!ASSERT(this._open, "Unable to open the file for writing!"))
        return false;

    var me = this;

    function writeProp(name, val)
    {
        me._fileStream.write(ecmaEscape(name) + " " + val + me.lineEnd);
    };

    function writeObjProps(o)
    {
        for (var p in o)
        {
            switch (typeof o[p])
            {
                case "string":
                    writeProp(p, o[p].quote());
                    break;

                case "number":
                    writeProp(p, o[p]);
                    break;

                case "function":
                    if (o[p] instanceof RegExp)
                        writeProp(p, o[p]);
                    // Can't serialize non-RegExp functions (yet).
                    break;

                case "null":
                    writeProp(p, "null");
                    break;

                case "undefined":
                    writeProp(p, "undefined");
                    break;

                case "object":
                    me._fileStream.write("START " + ecmaEscape(p) +
                                         this.lineEnd);
                    writeObjProps(o[p]);
                    me._fileStream.write("END" + this.lineEnd);
                    break;

                default:
                    // Can't handle anything else!
            }
        }
    };

    this._fileStream.write("START" + this.lineEnd);
    writeObjProps(obj);
    this._fileStream.write("END" + this.lineEnd);
}

/* deserialize()
 *
 * Reads in enough of the file to deserialize (realize) a single object. The
 * object deserialized is returned; all sub-properties of the object are
 * deserialized with it.
 */
TextSerializer.prototype.deserialize =
function ts_deserialize()
{
    if (!this._open)
        this.open("<");
    if (!ASSERT(this._open, "Unable to open the file for reading!"))
        return false;

    var obj = null;
    var rv = null;
    var objs = new Array();

    while (true)
    {
        if (this._lines.length == 0)
        {
            this._buffer += this._fileStream.read();
            // Got more data in the buffer, so split into lines.
            // The last one doesn't count - the rest get added to the full list.
            var lines = this._buffer.split(/[\r\n]+/);
            this._buffer = lines.pop();
            this._lines = this._lines.concat(lines);
            if (this._lines.length == 0)
                break;
        }

        // Split each line into "command params...".
        var parts = this._lines[0].match(/^\s*(\w+)(?:\s+(.*))?$/);
        var command = parts[1].toLowerCase();
        var params = parts[2];

        // 'start' and 'end' commands are special.
        switch (command)
        {
            case "start":
                if (!params)
                {
                    if (rv)
                        return rv;
                    rv = obj = new Object();
                }
                else
                {
                    /* Create a new object level, store the reference on the
                     * parent, and set the new object as the current.
                     */
                    var n = new Object();
                    objs.push(obj);
                    obj[ecmaUnescape(params)] = n;
                    obj = n;
                }
                this._lines.shift();
                break;

            case "end":
                this._lines.shift();
                if (rv && (objs.length == 0))
                {
                    // We're done for the day.
                    return rv;
                }
                // Return the to previous object level.
                obj = objs.pop();
                if (!ASSERT(obj, "Waaa! no object level to return too!"))
                    return rv;
                break;

            default:
                this._lines.shift();
                if (!obj)
                {
                    /* If we find a line that is NOT starting a new object, and
                     * we don't have a current object, we need to store the data
                     * on a dummy that we will omit the header for.
                     */
                    rv = obj = new Object();
                    obj._noHeader = true;
                }
                if (params[0] == '"') // String
                {
                    obj[ecmaUnescape(command)] = params.substr(1, params.length - 2);
                }
                else if (params[0] == "/") // RegExp
                {
                    var p = params.match(/^\/(.*)\/(\w*)$/);
                    obj[ecmaUnescape(command)] = new RegExp(p[1], p[2]);
                }
                else if (params == "null") // null
                {
                    obj[ecmaUnescape(command)] = null;
                }
                else if (params == "undefined") // undefined
                {
                    obj[ecmaUnescape(command)] = undefined;
                }
                else // Number
                {
                    obj[ecmaUnescape(command)] = Number(params);
                }
                break;
        }
    }
    return null;
}
